@@ -8,11 +8,14 @@ from deposit.query.query import Query
 from deposit.utils.fnc_files import (
 	as_url, url_to_path, extract_filename, get_image_format, delete_stored, 
 	store_locally, update_image_filename, get_updated_local_url, open_url, 
-	clear_temp_dir, prune_local_files
+	clear_temp_dir, get_named_path, get_free_subfolder, get_unique_path, 
+	prune_local_files,
 )
 
 from collections import defaultdict
 from natsort import natsorted
+from itertools import product
+import shutil
 import os
 
 class Store(object):
@@ -23,10 +26,13 @@ class Store(object):
 		
 		self._datasource = Datasource.Memory()
 		self._resources = {}			# {url: DResource, ...}
+		self._added = set()				# set(path, ...)
+		self._deleted = set()			# set((filename, path), ...)
 		self._user_tools = []
 		self._queries = {}
 		self._local_folder = None		# folder to store DResource files flagged as is_stored
 		self._keep_temp = keep_temp
+		self._do_backup = True
 		self._saved = True
 		
 		self._min_free_id = None
@@ -34,6 +40,7 @@ class Store(object):
 		self._descriptors = None
 		self._relation_labels = None
 		
+		self._callbacks_paused = False
 		self._callback_added = set()
 		self._callback_deleted = set()
 		self._callback_changed = set()
@@ -42,6 +49,7 @@ class Store(object):
 		self._callback_local_folder_changed = set()
 		self._callback_user_tools_changed = set()
 		self._callback_queries_changed = set()
+		self._callback_settings_changed = set()
 		
 	def _init_min_free_id(self) -> None:
 		
@@ -84,6 +92,10 @@ class Store(object):
 	
 	# ---- Callbacks
 	# ------------------------------------------------------------------------
+	def set_callbacks_paused(self, state):
+		
+		self._callbacks_paused = state
+	
 	def set_callback_added(self, func):
 		# func(elements: list); elements = [DObject, DClass, ...]
 		
@@ -92,28 +104,38 @@ class Store(object):
 	def callback_added(self, elements):
 		
 		self._saved = False
+		if self._callbacks_paused:
+			return
 		for func in self._callback_added:
 			func(elements)
 	
 	def set_callback_deleted(self, func):
 		# func(elements: list); elements = [obj_id, name, ...]
 		
+		if self._callbacks_paused:
+			return
 		self._callback_deleted.add(func)
 	
 	def callback_deleted(self, elements):
 		
 		self._saved = False
+		if self._callbacks_paused:
+			return
 		for func in self._callback_deleted:
 			func(elements)
 	
 	def set_callback_changed(self, func):
 		# func(elements: list); elements = [DObject, DClass, ...]
 		
+		if self._callbacks_paused:
+			return
 		self._callback_changed.add(func)
 	
 	def callback_changed(self, elements):
 		
 		self._saved = False
+		if self._callbacks_paused:
+			return
 		for func in self._callback_changed:
 			func(elements)
 	
@@ -125,6 +147,8 @@ class Store(object):
 	def callback_saved(self, datasource):
 		
 		self._saved = True
+		if self._callbacks_paused:
+			return
 		for func in self._callback_saved:
 			func(datasource)
 	
@@ -136,6 +160,8 @@ class Store(object):
 	def callback_loaded(self):
 		
 		self._saved = True
+		if self._callbacks_paused:
+			return
 		for func in self._callback_loaded:
 			func()
 	
@@ -147,6 +173,8 @@ class Store(object):
 	def callback_local_folder_changed(self):
 		
 		self._saved = False
+		if self._callbacks_paused:
+			return
 		for func in self._callback_local_folder_changed:
 			func()
 	
@@ -158,6 +186,8 @@ class Store(object):
 	def callback_queries_changed(self):
 		
 		self._saved = False
+		if self._callbacks_paused:
+			return
 		for func in self._callback_queries_changed:
 			func()
 	
@@ -169,15 +199,30 @@ class Store(object):
 	def callback_user_tools_changed(self):
 		
 		self._saved = False
+		if self._callbacks_paused:
+			return
 		for func in self._callback_user_tools_changed:
 			func()
 	
+	def set_callback_settings_changed(self, func):
+		# func()
+		
+		self._callback_settings_changed.add(func)
+	
+	def callback_settings_changed(self):
+		
+		if self._callbacks_paused:
+			return
+		for func in self._callback_settings_changed:
+			func()
+	
 	
 	# ---- General
 	# ------------------------------------------------------------------------
 	def clear(self) -> None:
 		
 		self.G.clear()
+		self.flush_added_deleted()
 		
 		self._resources = {}
 		self._user_tools = []
@@ -190,6 +235,21 @@ class Store(object):
 		self._descriptors = None
 		self._relation_labels = None
 	
+	def flush_added_deleted(self):
+		
+		for path in self._added:
+			filename = os.path.basename(path)
+			tgt_path = os.path.join(get_named_path("_deleted", self.get_folder()), filename)
+			shutil.move(path, tgt_path)
+		
+		for filename, path in self._deleted:
+			filename = os.path.basename(path)
+			tgt_path = get_unique_path(filename, get_free_subfolder(self.get_folder()))
+			shutil.move(path, tgt_path)
+		
+		self._added = set()
+		self._deleted = set()
+	
 	def get_updated_url(self, resource):
 		
 		if not resource.is_stored:
@@ -207,14 +267,23 @@ class Store(object):
 			return
 		if not os.path.isdir(path):
 			os.makedirs(path, exist_ok = True)
+		prev_local_folder = self._local_folder
 		self._local_folder = os.path.normpath(os.path.abspath(path))
 		
 		cmax = len(self.get_objects())
 		cnt = 1
 		to_del = set()
+		added_urls = set()
 		for obj in self.get_objects():
 			if progress is not None:
+				if progress.cancel_pressed():
+					self._local_folder = prev_local_folder
+					for url_new in added_urls:
+						del self._resources[url_new]
+					progress.stop()
+					return
 				progress.update_state(text = "Copying Resources", value = cnt, maximum = cmax)
+				cnt += 1
 			for descr in obj.get_descriptors():
 				resource = obj.get_descriptor(descr)
 				if isinstance(resource, DResource):
@@ -226,6 +295,7 @@ class Store(object):
 						resource.value = (url_new, filename, is_stored, is_image)
 						self._resources[url_new] = resource
 						to_del.add(url)
+						added_urls.add(url_new)
 		for url in to_del:
 			if url in self._resources:
 				del self._resources[url]
@@ -259,10 +329,11 @@ class Store(object):
 		if folder is None:
 			return (url, False)
 		
-		url_new = store_locally(url, filename, folder, self.get_resource_urls())
-		if url_new is None:
+		path = store_locally(url, filename, folder, self.get_resource_urls())
+		if path is None:
 			return (url, False)
-		return (url_new, True)
+		self._added.add(path)
+		return (as_url(path), True)
 	
 	def prune_resources(self):
 		
@@ -318,7 +389,12 @@ class Store(object):
 		
 		return self.G.get_class_data(name)
 	
-	def add_resource(self, url: str, filename: str = None, is_stored: bool = None, is_image: bool = None) -> DResource:
+	def add_resource(self, 
+		url: str, 
+		filename: str = None, 
+		is_stored: bool = None, 
+		is_image: bool = None
+	) -> DResource:
 		# filename = "name.ext"; if None, determine automatically
 		# is_stored = True if resource is stored in local folder
 		#	= False, not stored locally
@@ -621,7 +697,7 @@ class Store(object):
 	
 	
 	def get_descriptors(self, ordered = False) -> list:
-		# return [DClass, ...]
+		# returns [DClass, ...]
 		
 		if self._descriptors is None:
 			self._descriptors = set()
@@ -637,10 +713,24 @@ class Store(object):
 		return list(self._descriptors)
 	
 	def get_descriptor_names(self, ordered = False) -> list:
-		# return [name, ...]
+		# returns [name, ...]
 		
 		return [cls.name for cls in self.get_descriptors(ordered)]
 	
+	def get_descriptor_values(self, class_name, descriptor_name, direct_only = True) -> list:
+		# direct_only = if True, don't return members of subclasses
+		#
+		# returns [value, ...]
+		
+		cls = self.get_class(class_name)
+		descr = self.get_class(descriptor_name)
+		if (cls is None) or (descr is None):
+			return []
+		for obj in cls.get_members(direct_only):
+			value = obj.get_descriptor(descr)
+			if value is not None:
+				yield value
+	
 	
 	def get_user_tools(self):
 		
@@ -671,18 +761,17 @@ class Store(object):
 		
 		return self._relation_labels
 	
+	def reverse_relation(self, label):
+		
+		if label.startswith("~"):
+			return label[1:]
+		return "~" + label
 	
-	def get_class_connections(self, classes: list, relations: list, progress = None) -> list:
+	def get_class_connections(self, classes, relations, progress = None):
 		# classes = [class_name, ...]
 		# relations = [(class1, label, class2), ...]
-		#
-		# returns [path, ...]; where path = [obj_id, ...]
-		# 	first obj in path is a member of class_name from the first select
-		# at least one class must be specified
-		# if relations are specified, only accept connections between 
-		# 	[class1] and [class2] if they are labeled [label]
 		
-		def get_class_members(class_name):
+		def _get_class_members(class_name, leave_names = False):
 			
 			if class_name == "*":
 				return set([obj.id for obj in self.get_objects()])
@@ -693,104 +782,246 @@ class Store(object):
 			cls = self.get_class(class_name)
 			if cls is None:
 				return set()
+			if leave_names:
+				return class_name
 			return set([obj.id for obj in cls.get_members()])
 		
-		def get_paths(src, objects2, relation_objs, n_relations, class_objs, n_classes, has_asterisk_label):
+		def _get_class_names(cls):
 			
-			if (not objects2) and (not relation_objs):
-				return set([(src,)])
+			if cls == "*":
+				return set([cls_.name for cls_ in self.get_classes()])
+			if cls == "!*":
+				return set(["!*"])
+			if isinstance(cls, tuple):
+				clss = set.union(*[set([cls_.name for cls_ in \
+					self.get_object(obj_id).get_classes()]) for \
+						obj_id in list(cls)])
+				if not clss:
+					clss = set(["!*"])
+				return clss
+			return set([cls])
+		
+		def _get_within_class_rel(cls1, cls2):
 			
-			queue = [[[src], set(), class_objs.get(src, set()).copy()]]  
-			# queue = [[path, found_relations, found_classes], ...]
+			clss1 = _get_class_names(cls1)
+			clss2 = _get_class_names(cls2)
+			return clss1.intersection(clss2)
+		
+		def _get_paths(
+			obj_id0, cls_lookup, rels, within_cls_rels, 
+			asterisk_rels, classes, connecting, mandatory_classes,
+		):
+			
+			'''
+			Get all paths from obj_id0 to objects from specified classes.
+			1. If classes or relations contains object specifications OBJ(1, ...)
+			require each path to contain one of those objects
+			2. If a cls1.*.cls2 relation is specified, the path must contain an 
+			edge between cls1 and cls2.
+			3. The paths cannot have within-class edges unless explicitly
+			specified in relations.
+			4. If a path contains an edge between classes as specified in 
+			relations, it must have a label as specified in relations.
+			'''
+			queue = [[[obj_id0], set(), cls_lookup[obj_id0].copy(), False, False]]
+			# queue = [[path, found_asterisk, found_classes, rule1, rule3], ...]
 			done = set()
+			skip = set()
 			while queue:
-				path, found_rels, found_clss = queue.pop()
-				found = False
-				src_clss = class_objs.get(path[-1], set())
-				for tgt, label in self.G.iter_object_relations(path[-1]):
-					if tgt in path:
-						continue
-					idx = relation_objs.get((path[-1], label, tgt), None) if relation_objs else 0
-					if (idx is None) and has_asterisk_label:
-						idx = relation_objs.get((path[-1], "*", tgt), None)
-					if idx is None:
-						continue
-					tgt_clss = class_objs.get(tgt, set())
-					found_clss.update(tgt_clss)
-					is_same_class = (len(tgt_clss.intersection(src_clss)) > 0)
-					if len(found_clss) >= n_classes:
-						done.add(tuple(sorted(path + ([] if is_same_class else [tgt]))))
-					else:
-						found_rels.add(idx)
-						queue.append([path + [tgt], found_rels, found_clss])
-					found = True
-				if (not found) and (len(found_rels) >= n_relations):
-					done.add(tuple(sorted(path)))
-			
+				path, found_asterisk, found_classes, rule1, rule3 = queue.pop()
+				found_next = False
+				if not mandatory_classes:
+					rule1 = True
+				if (len(found_classes) < len(classes)) or within_cls_rels:
+					
+					src = path[-1]
+					clss_src = cls_lookup.get(src, None)
+					
+					# rule 1 - first stage, source
+					if (not rule1) and (clss_src is not None) and \
+						clss_src.intersection(mandatory_classes):
+							rule1 = True
+					
+					for tgt, label in self.G.iter_object_relations(src):
+						
+						# already processed
+						if tgt in skip:
+							continue
+						skip.add(tgt)
+						
+						# circular path
+						if tgt in path:
+							continue
+						
+						clss_tgt = cls_lookup.get(tgt, None)
+						
+						# only check objects with specified classes or from the connecting set
+						if (clss_tgt is None) and (tgt not in connecting):
+							continue
+						
+						# collect all combinations of source & target classes
+						clss = []
+						if (clss_src is not None) and (clss_tgt is not None):
+							if rels or asterisk_rels:
+								clss = list(product(clss_src, clss_tgt))
+						
+						# rule 2 - first stage
+						if clss and asterisk_rels:
+							for cls_src, cls_tgt in clss:
+								if (cls_src, cls_tgt) in asterisk_rels:
+									found_asterisk.add((cls_src, cls_tgt))
+						
+						# rule 3 - first stage
+						found_classes_ = found_classes.copy()
+						if clss_tgt is not None:
+							common = found_classes_.intersection(clss_tgt)
+							if common:
+								if (not rule3) and (clss_src is not None):
+									for cls in clss_src:
+										if (cls in within_cls_rels) and (within_cls_rels[cls].intersection([label, "*"])):
+											rule3 = True
+											break
+								continue
+						
+						# rule 1 - first stage, target
+						if clss_tgt is not None:
+							if (not rule1) and clss_tgt.intersection(mandatory_classes):
+								rule1 = True
+						
+						# rule 4
+						if clss:
+							found_rel = False
+							found_label = False
+							if rels:
+								for cls_src, cls_tgt in clss:
+									if (cls_src, cls_tgt) in rels:
+										found_rel = True
+										if rels[(cls_src, cls_tgt)].intersection(
+											[label, "*"]
+										) or rels[(cls_tgt, cls_src)].intersection(
+											[self.reverse_relation(label), "*"]
+										):
+											found_label = True
+											break
+							if found_rel and not found_label:
+								continue
+						
+						if clss_tgt is not None:
+							found_classes_.update(clss_tgt)
+						
+						queue.append([path + [tgt], found_asterisk, found_classes_, rule1, rule3])
+						found_next = True
+				
+				if found_next:
+					continue
+				
+				# rule 2 - second stage
+				rule2 = True
+				for cls1, cls2 in asterisk_rels:
+					if not (((cls1, cls2) in found_asterisk) or \
+						((cls2, cls1) in found_asterisk)):
+							rule2 = False
+							break
+				if not rule2:
+					continue
+				
+				# rule 3 - second stage
+				if within_cls_rels and not rule3:
+					continue
+				
+				# rule 1 - second stage
+				if mandatory_classes and not rule1:
+					continue
+				
+				# only add objects from specified classes
+				done.add(tuple(sorted(
+					[obj_id for obj_id in path if obj_id in cls_lookup]
+				)))
+				
 			return done
 		
-		relation_objs = {}
-		has_asterisk_label = False
-		for idx in range(len(relations)):
-			cls1, label, cls2 = relations[idx]
-			if (cls1, label, cls2) == ("*","*","*"):
-				continue
-			objects1 = get_class_members(cls1)
-			objects2 = get_class_members(cls2)
-			if label.startswith("~"):
-				rev_label = label[1:]
-			elif label == "*":
-				has_asterisk_label = True
-				rev_label = "*"
+		rels = defaultdict(set)  # {(class1, class2): set(label, ...), ...}
+		within_cls_rels = defaultdict(set)  # {class: set(label, ...), ...}
+		asterisk_rels = set()  # set((class1, class2), ...)
+		mandatory_classes = set() # set(class, ...)
+		for cls1, lbl, cls2 in relations:
+			if isinstance(cls1, tuple):
+				mandatory_classes.add(cls1)
+			if isinstance(cls2, tuple):
+				mandatory_classes.add(cls2)
+			if lbl == "*":
+				asterisk_rels.add((cls1, cls2))
+				asterisk_rels.add((cls2, cls1))
+			clss = _get_within_class_rel(cls1, cls2)
+			if clss:
+				within_cls_rels[cls1].add(lbl)
 			else:
-				rev_label = "~" + label
-			for src in objects1:
-				for tgt in objects2:
-					relation_objs[(src, label, tgt)] = idx
-					relation_objs[(tgt, rev_label, src)] = idx
-		n_relations = len(relations)
+				rels[(cls1, cls2)].add(lbl)
+				rels[(cls2, cls1)].add(self.reverse_relation(lbl))
 		
-		classes = [class_name for class_name in classes if \
-			(self.has_class(class_name) or \
-			isinstance(class_name, tuple) or \
-			(class_name in ["*", "!*"]))]
-		class_objs = defaultdict(set)
-		for class_name in classes:
-			for obj_id in get_class_members(class_name):
-				class_objs[obj_id].add(class_name)
-		n_classes = len(classes)
+		cls0 = classes[0]
+		objects0 = _get_class_members(cls0)
+		if not objects0:
+			# no objects found in the first specified class
+			return set()
+		
+		cls_lookup = defaultdict(set)
+		collect = []
+		for cls in classes:
+			if isinstance(cls, tuple):
+				mandatory_classes.add(cls)
+			objs = objects0 if (cls == cls0) else _get_class_members(cls)
+			if objs:
+				collect.append(cls)
+				for obj_id in objs:
+					cls_lookup[obj_id].add(cls)
+		classes = collect
+		
+		if (len(classes) == 1) and (classes[0] not in within_cls_rels):
+			# only one class and no within-class relations apply
+			return set([(obj_id,) for obj_id in objects0])
+		
+		connecting = set()
+		cls0_ = _get_class_members(cls0, leave_names = True)
+		if len(classes) > 1:
+			# collect objects from unspecified classes at shortest paths between specified classes
+			for cls1 in classes[1:]:
+				if (cls0, cls1) in rels:
+					continue
+				cls1_ = _get_class_members(cls1, leave_names = True)
+				connecting.update(self.G.shortest_path_between_classes(cls0_, cls1_))
+			connecting = connecting.difference(classes)
+			if connecting:
+				connecting = set.union(*[_get_class_members(cls) for cls in connecting])
 		
 		paths = set()
-		objects1 = get_class_members(classes[0])
-		objects2 = set()
-		for class_name in classes[1:]:
-			objects2.update(get_class_members(class_name))
-		cmax = len(objects1)
+		cmax = len(objects0)
 		cnt = 1
-		for obj_id in objects1:
-			if (progress is not None) and (cnt % 1000 == 0):
+		if progress is not None:
+			progress.update_state(value = 0, maximum = cmax)
+		for obj_id0 in objects0:
+#			print("\r%d/%d           " % (cnt, cmax), end = "") # DEBUG
+			if (progress is not None) and (cnt % 100 == 0):
 				if progress.cancel_pressed():
 					return []
-				progress.update_state(value = cnt, maximum = cmax)
+				progress.update_state(value = cnt)
 			cnt += 1
-			found = False
-			paths_ = get_paths(obj_id, objects2, relation_objs, n_relations, class_objs, n_classes, has_asterisk_label)
+			paths_ = _get_paths(
+				obj_id0, cls_lookup, rels, within_cls_rels, 
+				asterisk_rels, classes, connecting, mandatory_classes,
+			)
 			if paths_:
 				paths.update(paths_)
-			elif not relations:
-				paths.add((obj_id,))
 		
 		if progress is not None:
-			progress.update_state(value = cmax, maximum = cmax)
+			progress.update_state(value = cmax)
 		
 		return paths
 	
-	
 	def get_query(self, querystr: str, **kwargs) -> object:
 		
 		return Query(self, querystr, **kwargs)
 	
-	
 	def get_subgraph(self, objects: list, progress = None):
 		# returns store which contains specified objects with their descriptors
 		# and objects related to them, except objects from the same class as the
@@ -904,7 +1135,10 @@ class Store(object):
 			obj_id = obj_id.id
 		if not self.G.has_object(obj_id):
 			return
-		if self.get_object(obj_id).has_descriptors():
+		obj = self.get_object(obj_id)
+		if obj.has_descriptors():
+			for descr in obj.get_descriptors():
+				obj.del_descriptor(descr)
 			self._descriptors = None
 		self.G.del_object(obj_id)
 		self._init_min_free_id()
@@ -939,7 +1173,9 @@ class Store(object):
 	def del_resource(self, resource: DResource) -> None:
 		
 		if resource.is_stored:
-			delete_stored(resource.url, self.get_folder())
+			filename, path = delete_stored(resource.url, self.get_folder())
+			if path is not None:
+				self._deleted.add((filename, path))
 		url = resource.url
 		if url in self._resources:
 			del self._resources[url]
@@ -1007,11 +1243,29 @@ class Store(object):
 		
 		self._datasource = datasource
 	
+	def has_auto_backup(self):
+		
+		return self._do_backup
+	
+	def set_auto_backup(self, state):
+		
+		self._do_backup = state
+		self.callback_settings_changed()
+	
 	def save(self, datasource = None, *args, **kwargs):
 		
+		if self._do_backup and (self._datasource is not None):
+			folder = self._datasource.get_folder()
+			if folder is None:
+				folder = self.get_folder()
+			folder = get_named_path("_backup", folder)
+			self._datasource.backup(self, folder)
+		
 		# datasource = Datasource or format
 		datasource = self.init_datasource(datasource, kwargs)
 		if datasource.save(self, datasource = datasource, *args, **kwargs):
+			self._added = set()
+			self._deleted = set()
 			self.callback_saved(datasource)
 			return True
 		return False
@@ -1020,9 +1274,12 @@ class Store(object):
 		# datasource = Datasource or format
 		
 		datasource = self.init_datasource(datasource, kwargs)
+		self.set_callbacks_paused(True)
 		if datasource.load(self, datasource = datasource, *args, **kwargs):
+			self.set_callbacks_paused(False)
 			self.callback_loaded()
 			return True
+		self.set_callbacks_paused(False)
 		return False
 	
 	def is_saved(self):
